
/*
 * Original code /usr/src/crypto/heimdal/lib/roken/base64.c on my computer
 * from Kungliga Tekniska Hgskolan (Royal Institute of Technology,
 * Stockholm, Sweden)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// open, write, chmod
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/stat.h>

#include "LsmBase64.h"
#include "LsmTypes.h"

static char     impl_Base64_chars[] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static int
impl_Base64_pos(char c)
{
	char           *p;
	for (p = impl_Base64_chars; *p; p++)
		if (*p == c)
			return p - impl_Base64_chars;
	return -1;
}

static int
impl_Base64_encode(const void *data,
	      int size,
	      char **str)
{
	char           *s, *p;
	int             i;
	int             c;
	const unsigned char *q;

	p = s = (char *)malloc(size * 4 / 3 + 4);
	if (p == NULL)
		return -1;
	q = (const unsigned char *)data;
	i = 0;
	for (i = 0; i < size;) {
		c = q[i++];
		c *= 256;
		if (i < size)
			c += q[i];
		i++;
		c *= 256;
		if (i < size)
			c += q[i];
		i++;
		p[0] = impl_Base64_chars[(c & 0x00fc0000) >> 18];
		p[1] = impl_Base64_chars[(c & 0x0003f000) >> 12];
		p[2] = impl_Base64_chars[(c & 0x00000fc0) >> 6];
		p[3] = impl_Base64_chars[(c & 0x0000003f) >> 0];
		if (i > size)
			p[3] = '=';
		if (i > size + 1)
			p[2] = '=';
		p += 4;
	}
	*p = 0;
	*str = s;
	return strlen(s);
}

#define DECODE_ERROR 0xffffffff

static unsigned int
impl_Base64_tokenDecode(const char *token)
{
	int             i;
	unsigned int    val = 0;
	int             marker = 0;
	if (strlen(token) < 4)
		return DECODE_ERROR;
	for (i = 0; i < 4; i++) {
		val *= 64;
		if (token[i] == '=')
			marker++;
		else if (marker > 0)
			return DECODE_ERROR;
		else
			val += impl_Base64_pos(token[i]);
	}
	if (marker > 2)
		return DECODE_ERROR;
	return (marker << 24) | val;
}

static int
impl_Base64_decode(const char *str,
	      void *data)
{
	const char     *p;
	unsigned char  *q;

	q = data;
	for (p = str; *p && (*p == '=' || strchr(impl_Base64_chars, *p)); p += 4) {
		unsigned int    val = impl_Base64_tokenDecode(p);
		unsigned int    marker = (val >> 24) & 0xff;
		if (val == DECODE_ERROR)
			return -1;
		*q++ = (val >> 16) & 0xff;
		if (marker < 2)
			*q++ = (val >> 8) & 0xff;
		if (marker < 1)
			*q++ = val & 0xff;
	}
	return q - (unsigned char *)data;
}

int
LsmBase64_encodeFromFile(char *in_file,
			  char **out_dest,
			  int *out_destSize)
{
	int             retval = 0;
	GError         *error=NULL;
	int             inLen, outLen;
	char           *inStr;

	g_return_val_if_fail(in_file, LSM__ERROR__NULL_POINTER);
	g_return_val_if_fail(out_dest, LSM__ERROR__NULL_POINTER);
	g_return_val_if_fail(g_file_test(in_file, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_EXISTS),
			     LSM__ERROR__FILE_NOT_FOUND);

	g_file_get_contents(in_file, &inStr, &inLen, &error);
	outLen = impl_Base64_encode(inStr, inLen, out_dest);
	g_return_val_if_fail(outLen!=-1, LSM__ERROR__NO_MEMORY);
	if(out_destSize)
		*out_destSize=outLen;
	g_free(inStr);

	return retval;
}

int
LsmBase64_decodeToFile(char *in_src,
			char *in_file)
{
	int             retval = 0;
	int             outLen;
	int             fd;

	char           *dest;

	g_return_val_if_fail(in_src, LSM__ERROR__NULL_POINTER);
	g_return_val_if_fail(in_file, LSM__ERROR__NULL_POINTER);
	
	dest = g_new0(char, strlen(in_src));
	g_return_val_if_fail(dest, LSM__ERROR__NO_MEMORY);

	outLen = impl_Base64_decode(in_src, dest);

	g_return_val_if_fail(outLen!=-1, LSM__ERROR__BASE64_DECODE);

	fd = open(in_file, O_WRONLY | O_CREAT);
	g_return_val_if_fail(fd!=-1, LSM__ERROR__INVALID_FILENAME);
	write(fd, dest, outLen);
	close(fd);
	chmod(in_file, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

	g_free(dest);

	return retval;
}

int
LsmBase64_encodeFromStream(char *in_stream,
		int in_streamLength,
		char **out_dest,
		int *out_destSize)
{
	int retval=0;
	int destSize;
	int streamLength;

	g_return_val_if_fail(in_stream, LSM__ERROR__NULL_POINTER);
	g_return_val_if_fail(out_dest, LSM__ERROR__NULL_POINTER);

	if (in_streamLength)
		streamLength=in_streamLength;
	else	
		streamLength=strlen(in_stream);
	
	destSize=impl_Base64_encode(in_stream, streamLength, out_dest);
	g_return_val_if_fail(destSize!=-1, LSM__ERROR__NO_MEMORY);
	if(out_destSize)
		*out_destSize=destSize;

	return retval;
}

int
LsmBase64_decodeToStream(char *in_src,
		char **out_stream,
		int *out_streamLength)
{
	int retval=0;
	int             outLen;
	char           *dest;

	g_return_val_if_fail(in_src, LSM__ERROR__NULL_POINTER);
	g_return_val_if_fail(out_stream, LSM__ERROR__NULL_POINTER);

	dest = g_new0(char, strlen(in_src)*2/3+1);
	g_return_val_if_fail(dest, LSM__ERROR__NO_MEMORY);

	outLen = impl_Base64_decode(in_src, dest);
	
	g_return_val_if_fail(outLen!=-1, LSM__ERROR__BASE64_DECODE);

	*out_stream=dest;
	
	if(out_streamLength)
		*out_streamLength=outLen;

	return retval;
}
	

